from typing import Optional
from django.contrib.auth import get_user_model
# strawberry
import strawberry
from strawberry.django import auth
from gqlauth.utils import g_user
# project
from .types import UserType, UserFilter
from typing import List



@strawberry.type
class UserQueries:
    user: Optional[UserType] = strawberry.django.field()
    users: UserType = strawberry.django.field(filters=UserFilter)
    # me: UserType = strawberry.django.field(resolver=lambda info: g_user(info))
    #
    # @strawberry.field
    # def public_user(self, info) -> Optional[UserType]:
    #     user = g_user(info)
    #     if user.is_authenticated:
    #         return user
    #     return None



from dataclasses import dataclass
from base64 import b64decode, b64encode
import io
from gqlauth.settings import gqlauth_settings as app_settings
from pathlib import Path
import sys
from PIL.Image import Image
from faker import Faker
from .create import ImageCaptcha
sys.path.append(str(Path(__file__).parent.parent.parent))
FONTS_PATH = str(Path(__file__).parent.joinpath('fonts'))
fake = Faker()


@dataclass
class CaptchaType:
    image: Image
    text: str

    def as_base64(self):
        bytes_array = io.BytesIO()
        self.image.save(bytes_array,format='PNG')
        return b64encode(bytes_array.getvalue())

    def show(self):
        self.image.show()


def get_image(text):
    image = ImageCaptcha(width=300,
    height=150,
    heb_fonts=[FONTS_PATH+'/stam.ttf'],
    fonts=[FONTS_PATH+'/OpenSans-Semibold.ttf'],
    )
    image = image.generate_image(text)
    return image


def generate_text() -> str:
    factory = app_settings.CAPTCHA_TEXT_FACTORY
    if factory:
        return factory()
    return " ".join([fake.city(), str(fake.pyint())])


def generate_city_captcha():
    text = generate_text()
    image = get_image(text)
    return CaptchaType(image=image, text=text)




import warnings
import inspect
from django.core import signing
from django.conf import settings as django_settings
from strawberry.annotation import StrawberryAnnotation
from strawberry.utils.str_converters import to_camel_case
from strawberry.arguments import StrawberryArgument, UNSET
from .exceptions import TokenScopeError


def hide_args_kwargs(field):
    sig = inspect.signature(field)
    cleared = tuple(
        p for p in sig.parameters.values() if p.name not in ('kwargs', 'args')
    )
    field.__signature__ = inspect.signature(field).replace(parameters=(cleared))
    return field

def isiterable(value):
    try:
        iter(value)
    except TypeError:
        return False
    return True

def camelize(data):
    if isinstance(data, dict):
        return {to_camel_case(k): camelize(v) for k, v in data.items()}
    if isiterable(data) and not isinstance(data, str):
        return [camelize(d) for d in data]
    return data

def list_to_dict(lst: [str]):
    """takes list of string and creates a dict with str as their values"""
    new_dict = {}
    for item in lst:
        new_dict[item] = str
    return new_dict

warnings.simplefilter("once")


def g_user(info):
    # returns a user from info obj
    user = getattr(info.context, 'user', False)
    if user: return user
    return info.context.request.user


def get_token(user, action, **kwargs):
    username = user.get_username()
    if hasattr(username, "pk"):
        username = username.pk
    payload = {user.USERNAME_FIELD: username, "action": action}
    if kwargs:
        payload.update(**kwargs)
    token = signing.dumps(payload)
    return token


def get_payload_from_token(token, action, exp=None):
    payload = signing.loads(token, max_age=exp)
    _action = payload.pop("action")
    if _action != action:
        raise TokenScopeError
    return payload



def using_refresh_tokens():
    if (
            hasattr(django_settings, "GRAPHQL_JWT")
            and django_settings.GRAPHQL_JWT.get("JWT_LONG_RUNNING_REFRESH_TOKEN", False)
            and "strawberry_django_jwt.refresh_token"
            in django_settings.INSTALLED_APPS
    ):
        return True
    return False


def revoke_user_refresh_token(user):
    if using_refresh_tokens():
        refresh_tokens = user.refresh_tokens.all()
        for refresh_token in refresh_tokens:
            try:
                refresh_token.revoke()
            except Exception:  # JSONWebTokenError
                pass


def flat_dict(dict_or_list):
    """
    if is dict, return list of dict keys,
    if is list, return the list
    """
    return list(dict_or_list.keys()) if isinstance(dict_or_list, dict) else dict_or_list


def normalize_fields(dict_or_list, extra_list_or_dict):
    """
    helper merge settings defined fileds
    with default str type
    """
    if not isinstance(extra_list_or_dict, dict):
        extra_list_or_dict = list_to_dict(extra_list_or_dict)
    if not isinstance(dict_or_list, dict):
        dict_or_list = list_to_dict(dict_or_list)
    dict_or_list.update(extra_list_or_dict)

    return dict_or_list






def create_strawberry_argument(
        python_name: str, graphql_name: str, type_, default=None
):
    return StrawberryArgument(
        python_name=python_name,
        graphql_name=graphql_name,
        type_annotation=StrawberryAnnotation(type_),
        default=default or UNSET
    )



from smtplib import SMTPException
from uuid import UUID

from django.core.signing import BadSignature, SignatureExpired
from django.core.exceptions import ObjectDoesNotExist
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import SetPasswordForm, PasswordChangeForm
from django.db import transaction
from django.utils.module_loading import import_string

from strawberry_django_jwt.exceptions import JSONWebTokenError, JSONWebTokenExpired

from gqlauth.forms import RegisterForm, EmailForm, UpdateAccountForm, PasswordLessRegisterForm
from gqlauth.utils import normalize_fields, g_user
from gqlauth.models import Captcha, UserStatus
from gqlauth.settings import gqlauth_settings as app_settings
from gqlauth.exceptions import (
    UserAlreadyVerified,
    UserNotVerified,
    TokenScopeError,
    EmailAlreadyInUse,
    InvalidCredentials,
    PasswordAlreadySetError,
)
from gqlauth.constants import Messages, TokenAction
from gqlauth.utils import revoke_user_refresh_token, get_payload_from_token, using_refresh_tokens
from gqlauth.shortcuts import get_user_by_email, get_user_to_login
from gqlauth.signals import user_registered, user_verified
from gqlauth.decorators import (
    password_confirmation_required,
    verification_required,
    secondary_email_required,
    login_required
)

UserModel = get_user_model()
if app_settings.EMAIL_ASYNC_TASK and isinstance(app_settings.EMAIL_ASYNC_TASK, str):
    async_email_func = import_string(app_settings.EMAIL_ASYNC_TASK)
else:
    async_email_func = None


class RegisterMixin:
    """
    Register user with fields defined in the settings.

    If the email field of the user model is part of the
    registration fields (default), check if there is
    no user with that email or as a secondary email.

    If it exists, it does not register the user,
    even if the email field is not defined as unique
    (default of the default django user model).

    When creating the user, it also creates a `UserStatus`
    related to that user, making it possible to track
    if the user is archived, verified and has a secondary
    email.

    Send account verification email.

    If allowed to not verified users login, return token.
    """

    class _meta:

        password_fields = (
            []
            if app_settings.ALLOW_PASSWORDLESS_REGISTRATION
            else ["password1", "password2"]
        )
        extra_fields = password_fields
        if app_settings.REGISTER_REQUIRE_CAPTCHA:
            captcha_fields = {'identifier': UUID, "userEntry": str}
            extra_fields = normalize_fields(captcha_fields, password_fields)

        _required_inputs = normalize_fields(
            app_settings.REGISTER_MUTATION_FIELDS, extra_fields
        )
        _inputs = app_settings.REGISTER_MUTATION_FIELDS_OPTIONAL

    form = (
        PasswordLessRegisterForm
        if app_settings.ALLOW_PASSWORDLESS_REGISTRATION
        else RegisterForm
    )

    @classmethod
    def check_captcha(cls, input_):
        uuid = input_.get('identifier')
        try:
            obj = Captcha.objects.get(uuid=uuid)
        except Captcha.DoesNotExist:
            return Messages.CAPTCHA_EXPIRED
        return obj.validate(input_.get('userEntry'))

    @classmethod
    def resolve_mutation(cls, info, **input_):
        if app_settings.LOGIN_REQUIRE_CAPTCHA:
            check_res = cls.check_captcha(input_)
            if not check_res == Messages.CAPTCHA_VALID:
                return cls.output(success=False, errors={'captcha': check_res})

        try:
            with transaction.atomic():
                f = cls.form(input_)
                if f.is_valid():
                    email = input_.get('email')
                    UserStatus.clean_email(email)
                    user = f.save()
                    send_activation = (
                            app_settings.SEND_ACTIVATION_EMAIL is True and email
                    )
                    send_password_set = (
                            app_settings.ALLOW_PASSWORDLESS_REGISTRATION is True
                            and app_settings.SEND_PASSWORD_SET_EMAIL is True
                            and email
                    )
                    if send_activation:
                        # TODO CHECK FOR EMAIL ASYNC SETTING
                        if async_email_func:
                            async_email_func(user.status.send_activation_email, (info,))
                        else:
                            user.status.send_activation_email(info)

                    if send_password_set:
                        # TODO CHECK FOR EMAIL ASYNC SETTING
                        if async_email_func:
                            async_email_func(
                                user.status.send_password_set_email, (info,)
                            )
                        else:
                            user.status.send_password_set_email(info)

                    user_registered.send(sender=cls, user=user)
                    return cls.output(success=True)
                else:
                    return cls.output(success=False, errors=f.errors.get_json_data())
        except EmailAlreadyInUse:
            return cls.output(
                success=False,
                # if the email was set as a secondary email,
                # the RegisterForm will not catch it,
                # so we need to run UserStatus.clean_email(email)
                errors={UserModel.EMAIL_FIELD: Messages.EMAIL_IN_USE},
            )
        except SMTPException:
            return cls.output(success=False, errors=Messages.EMAIL_FAIL)


class VerifyAccountMixin:
    """
    Verify user account.

    Receive the token that was sent by email.
    If the token is valid, make the user verified
    by making the `user.status.verified` field true.
    """

    class _meta:
        _required_inputs = ["token"]

    @classmethod
    def resolve_mutation(cls, info, **input_):
        try:
            token = input_.get("token")
            UserStatus.verify(token)
            return cls.output(success=True)
        except UserAlreadyVerified:
            return cls.output(success=False, errors=Messages.ALREADY_VERIFIED)
        except SignatureExpired:
            return cls.output(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return cls.output(success=False, errors=Messages.INVALID_TOKEN)


class VerifySecondaryEmailMixin:
    """
    Verify user secondary email.

    Receive the token that was sent by email.
    User is already verified when using this mutation.

    If the token is valid, add the secondary email
    to `user.status.secondary_email` field.

    Note that until the secondary email is verified,
    it has not been saved anywhere beyond the token,
    so it can still be used to create a new account.
    After being verified, it will no longer be available.
    """

    class _meta:
        _required_inputs = ["token"]

    @classmethod
    def resolve_mutation(cls, info, **input_):
        try:
            token = input_.get("token")
            UserStatus.verify_secondary_email(token)
            return cls.output(success=True)
        except EmailAlreadyInUse:
            # while the token was sent and the user haven't
            # verified, the email was free. If other account
            # was created with it, it is already in use.
            return cls.output(success=False, errors=Messages.EMAIL_IN_USE)
        except SignatureExpired:
            return cls.output(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return cls.output(success=False, errors=Messages.INVALID_TOKEN)


class ResendActivationEmailMixin:
    """
    Sends activation email.

    It is called resend because theoretically
    the first activation email was sent when
    the user registered.

    If there is no user with the requested email,
    a successful response is returned.
    """

    class _meta:
        _required_inputs = ["email"]

    @classmethod
    def resolve_mutation(cls, info, **input_):
        try:
            email = input_.get("email")
            f = EmailForm({"email": email})
            if f.is_valid():
                user = get_user_by_email(email)
                if async_email_func:
                    async_email_func(user.status.resend_activation_email, (info,))
                else:
                    user.status.resend_activation_email(info)
                return cls.output(success=True)
            return cls.output(success=False, errors=f.errors.get_json_data())
        except ObjectDoesNotExist:
            return cls.output(success=True)  # even if user is not registred
        except SMTPException:
            return cls.output(success=False, errors=Messages.EMAIL_FAIL)
        except UserAlreadyVerified:
            return cls.output(success=False, errors={"email": Messages.ALREADY_VERIFIED})


class SendPasswordResetEmailMixin:
    """
    Send password reset email.

    For non verified users, send an activation
    email instead.

    Accepts both primary and secondary email.

    If there is no user with the requested email,
    a successful response is returned.
    """

    class _meta:
        _required_inputs = ["email"]

    @classmethod
    def resolve_mutation(cls, info, **input_):
        try:
            email = input_.get("email")
            f = EmailForm({"email": email})
            if f.is_valid():
                user = get_user_by_email(email)
                if async_email_func:
                    async_email_func(
                        user.status.send_password_reset_email, (info, [email])
                    )
                else:
                    user.status.send_password_reset_email(info, [email])
                return cls.output(success=True)
            return cls.output(success=False, errors=f.errors.get_json_data())
        except ObjectDoesNotExist:
            return cls.output(success=True)  # even if user is not registred
        except SMTPException:
            return cls.output(success=False, errors=Messages.EMAIL_FAIL)
        except UserNotVerified:
            user = get_user_by_email(email)
            try:
                if async_email_func:
                    async_email_func(user.status.resend_activation_email, (info,))
                else:
                    user.status.resend_activation_email(info)
                return cls(
                    success=False,
                    errors={"email": Messages.NOT_VERIFIED_PASSWORD_RESET},
                )
            except SMTPException:
                return cls.output(success=False, errors=Messages.EMAIL_FAIL)


class PasswordResetMixin:
    """
    Change user password without old password.

    Receive the token that was sent by email.

    If token and new passwords are valid, update
    user password and in case of using refresh
    tokens, revoke all of them.

    Also, if user has not been verified yet, verify it.
    """

    class _meta:
        _required_inputs = ["token", "new_password1", "new_password2"]

    form = SetPasswordForm

    @classmethod
    def resolve_mutation(cls, info, **input_):
        try:
            token = input_.get("token")
            payload = get_payload_from_token(
                token,
                TokenAction.PASSWORD_RESET,
                app_settings.EXPIRATION_PASSWORD_RESET_TOKEN,
            )
            user = UserModel._default_manager.get(**payload)
            form_dict = {
                'new_password1': input_['newPassword1'],
                'new_password2': input_['newPassword2']
            }
            f = cls.form(user, form_dict)
            if f.is_valid():
                revoke_user_refresh_token(user)
                user = f.save()
                if user.status.verified is False:
                    user.status.verified = True
                    user.status.save(update_fields=["verified"])
                    user_verified.send(sender=cls, user=user)

                return cls.output(success=True)
            return cls.output(success=False, errors=f.errors.get_json_data())
        except SignatureExpired:
            return cls.output(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return cls.output(success=False, errors=Messages.INVALID_TOKEN)


class PasswordSetMixin:
    """
    Set user password - for passwordless registration

    Receive the token that was sent by email.

    If token and new passwords are valid, set
    user password and in case of using refresh
    tokens, revoke all of them.

    Also, if user has not been verified yet, verify it.
    """

    class _meta:
        _required_inputs = ["token", "new_password1", "new_password2"]

    form = SetPasswordForm

    @classmethod
    def resolve_mutation(cls, info, **input_):
        try:
            token = input_.get("token")
            payload = get_payload_from_token(
                token,
                TokenAction.PASSWORD_SET,
                app_settings.EXPIRATION_PASSWORD_SET_TOKEN,
            )
            user = UserModel._default_manager.get(**payload)
            form_dict = {
                'new_password1': input_['newPassword1'],
                'new_password2': input_['newPassword2']
            }
            f = cls.form(user, form_dict)
            if f.is_valid():
                # Check if user has already set a password
                if user.has_usable_password():
                    raise PasswordAlreadySetError
                revoke_user_refresh_token(user)
                user = f.save()

                if user.status.verified is False:
                    user.status.verified = True
                    user.status.save(update_fields=["verified"])

                return cls.output(success=True)
            return cls.output(success=False, errors=f.errors.get_json_data())
        except SignatureExpired:
            return cls.output(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return cls.output(success=False, errors=Messages.INVALID_TOKEN)
        except (PasswordAlreadySetError):
            return cls.output(success=False, errors=Messages.PASSWORD_ALREADY_SET)


class ObtainJSONWebTokenMixin:
    """
    Obtain JSON web token for given user.

    Allow to perform login with different fields,
    and secondary email if set. The fields are
    defined on settings.

    Not verified users can login by default. This
    can be changes on settings.

    If user is archived, make it unarchive and
    return `unarchiving=True` on OutputBase.
    """

    class _meta:
        additional_req = {}
        if app_settings.LOGIN_REQUIRE_CAPTCHA:
            additional_req.update({'identifier': UUID, 'userEntry': str})
        _required_inputs = normalize_fields(app_settings.LOGIN_REQUIRED_FIELDS, additional_req)

        _inputs = app_settings.LOGIN_OPTIONAL_FIELDS
        _parent_resolver_name = 'obtain'

    @classmethod
    def check_captcha(cls, **input_):
        uuid = input_.get("identifier")
        try:
            obj = Captcha.objects.get(id=uuid)
        except Captcha.DoesNotExist:
            return Messages.CAPTCHA_EXPIRED

        return obj.validate(input_.get('userEntry'))

    @classmethod
    def resolve_mutation(cls, info, **input_):
        if app_settings.LOGIN_REQUIRE_CAPTCHA:
            check_res = cls.check_captcha(**input_)
            if not check_res == Messages.CAPTCHA_VALID:
                return cls.output(success=False, errors={'captcha': check_res})

        try:
            USERNAME_FIELD = UserModel.USERNAME_FIELD
            unarchiving = False

            # extract USERNAME_FIELD to use in query
            username = input_.get("username")
            password = input_.get("password")
            query_input_ = {USERNAME_FIELD: username}
            user = get_user_to_login(**query_input_)
            query_input_['password'] = password

            if user.status.archived is True:  # unarchive on login
                UserStatus.unarchive(user)
                unarchiving = True

            if user.status.verified or app_settings.ALLOW_LOGIN_NOT_VERIFIED:
                # this will raise if not successful
                res = cls.obtain.get_result(None, None, [cls, info], query_input_)

                return cls.output(success=True, obtainPayload=res)
            else:
                raise UserNotVerified

        except (JSONWebTokenError, ObjectDoesNotExist, InvalidCredentials):
            return cls.output(success=False, errors=Messages.INVALID_CREDENTIALS)
        except UserNotVerified:
            return cls.output(success=False, errors=Messages.NOT_VERIFIED)


class ArchiveOrDeleteMixin:
    @classmethod
    @verification_required
    @password_confirmation_required
    def resolve_mutation(cls, info, **input_):
        user = g_user(info)
        cls.resolve_action(user, info=info)
        return cls.output(success=True)


class ArchiveAccountMixin(ArchiveOrDeleteMixin):
    """
    Archive account and revoke refresh tokens.

    User must be verified and confirm password.
    """

    class _meta:
        _required_inputs = ["password"]

    @classmethod
    def resolve_action(cls, user, **input_):
        UserStatus.archive(user)
        revoke_user_refresh_token(user=user)


class DeleteAccountMixin(ArchiveOrDeleteMixin):
    """
    Delete account permanently or make `user.is_active=False`.

    The behavior is defined on settings.
    Anyway user refresh tokens are revoked.

    User must be verified and confirm password.
    """

    class _meta:
        _required_inputs = ["password"]

    @classmethod
    def resolve_action(cls, user, *args, **input_):
        if app_settings.ALLOW_DELETE_ACCOUNT:
            revoke_user_refresh_token(user=user)
            user.delete()


class PasswordChangeMixin:
    """
    Change account password when user knows the old password.

    A new token and refresh token are sent. User must be verified.
    """

    class _meta:
        _required_inputs = ["old_password", "new_password1", "new_password2"]
        _outputs = []
        if using_refresh_tokens():
            _outputs = ["refresh_token", 'token']
        _parent_resolver_name = 'obtain'

    form = PasswordChangeForm

    @classmethod
    @verification_required
    @password_confirmation_required
    def resolve_mutation(cls, info, **input_):
        user = g_user(info)
        form_dict = {
            'old_password': input_['oldPassword'],
            'new_password1': input_['newPassword1'],
            'new_password2': input_['newPassword2']
        }
        f = cls.form(user, form_dict)
        if f.is_valid():
            revoke_user_refresh_token(user)
            user = f.save()
            parent_input = {
                user.USERNAME_FIELD: getattr(user, user.USERNAME_FIELD),
                'password': input_.get("newPassword1")
            }
            parent_res = cls.obtain.get_result(
                None, None, [cls, info], parent_input
            )
            return cls.output(success=True, obtainPayload=parent_res)
        else:
            return cls.output(success=False, errors=f.errors.get_json_data())


class UpdateAccountMixin:
    """
    Update user model fields, defined on settings.

    User must be verified.
    """

    class _meta:
        _inputs = app_settings.UPDATE_MUTATION_FIELDS

    form = UpdateAccountForm

    @classmethod
    @verification_required
    def resolve_mutation(cls, info, **input_):
        user = g_user(info)
        f = cls.form(
        {
            'first_name': input_.get('firstName'),
            'last_name': input_.get('lastName')
        },
        instance=user)
        if f.is_valid():
            f.save()
            return cls.output(success=True)
        else:
            return cls.output(success=False, errors=f.errors.get_json_data())


class VerifyTokenMixin:
    """
    Checks if a token is not expired and correct
    """

    class _meta:
        _parent_resolver_name = 'verify'

    @classmethod
    def resolve_mutation(cls, info, **input_):
        try:
            payload = cls.verify.get_result(None, None, [cls, info], input_)
            return cls.output(success=True, verifyPayload=(payload))
        except JSONWebTokenExpired:
            return cls.output(success=False, errors=Messages.EXPIRED_TOKEN)
        except JSONWebTokenError:
            return cls.output(success=False, errors=Messages.INVALID_TOKEN)


class RefreshTokenMixin:
    """
    ### refreshToken to refresh your token:

    using the refresh token you already got during authorization.
    this will obtain a brand-new token (and possibly a refresh token)
    with renewed expiration time for non-expired tokens
    """

    class _meta:
        _parent_resolver_name = 'refresh'

    @classmethod
    def resolve_mutation(cls, info, **input_):
        try:
            parent_input = {
                'refresh_token': input_.get('refreshToken')
            }

            res = cls.refresh.get_result(None, None, [cls, info], parent_input)
            return cls.output(success=True, refreshPayload=res)

        except JSONWebTokenExpired:
            return cls.output(success=False, errors=Messages.EXPIRED_TOKEN)
        except JSONWebTokenError:
            return cls.output(success=False, errors=Messages.INVALID_TOKEN)


class RevokeTokenMixin:
    """
    Suspends a refresh token
    """

    class _meta:
        _parent_resolver_name = 'revoke'

    @classmethod
    def resolve_mutation(cls, info, **input_):
        try:
            parent_input = {
                'refresh_token': input_.get('refreshToken')
            }

            res = cls.revoke.get_result(None, None, [cls, info], parent_input)
            return cls.output(success=True, revokePayload=res)

        except JSONWebTokenExpired:
            return cls.output(success=False, errors=Messages.EXPIRED_TOKEN)
        except JSONWebTokenError:
            return cls.output(success=False, errors=Messages.INVALID_TOKEN)


class SendSecondaryEmailActivationMixin:
    """
    Send activation to secondary email.

    User must be verified and confirm password.
    """

    class _meta:
        _required_inputs = ["password"]

    @classmethod
    @verification_required
    @password_confirmation_required
    def resolve_mutation(cls, info, **input_):
        try:
            email = input_.get("email")
            f = EmailForm({"email": email})
            if f.is_valid():
                user = g_user(info)
                if async_email_func:
                    async_email_func(
                        user.status.send_secondary_email_activation, (info, email)
                    )
                else:
                    user.status.send_secondary_email_activation(info, email)
                return cls.output(success=True)
            return cls.output(success=False, errors=f.errors.get_json_data())
        except EmailAlreadyInUse:
            # while the token was sent and the user haven't verified,
            # the email was free. If other account was created with it
            # it is already in use
            return cls.output(success=False, errors={"email": Messages.EMAIL_IN_USE})
        except SMTPException:
            return cls.output(success=False, errors=Messages.EMAIL_FAIL)


class SwapEmailsMixin:
    """
    Swap between primary and secondary emails.

    Require password confirmation.
    """

    class _meta:
        _required_inputs = ["password"]

    @classmethod
    @secondary_email_required
    @password_confirmation_required
    def resolve_mutation(cls, info, **input_):
        info.context.user.status.swap_emails()
        return cls.output(success=True)


class RemoveSecondaryEmailMixin:
    """
    Remove user secondary email.

    Require password confirmation.
    """

    class _meta:
        _required_inputs = ["password"]

    @classmethod
    @secondary_email_required
    @password_confirmation_required
    def resolve_mutation(cls, info, **input_):
        info.context.user.status.remove_secondary_email()
        return cls.output(success=True)



import typing

import strawberry
from gqlauth.bases.scalars import ExpectedErrorType


@strawberry.interface
class OutputInterface:
    """
    A class to all public classes extend to
    padronize the output
    """
    success: bool
    errors: typing.Optional[ExpectedErrorType] = None

class MutationMixin:
    """
    All user should extend this class
    """

    @classmethod
    def mutate(cls, root, info, **input):
        return cls.resolve_mutation(root, info, **input)

    @classmethod
    def parent_resolve(cls, root, info, **kwargs):
        return super().mutate(root, info, **kwargs)


class DynamicArgsMixin:
    """
    A class that knows how to initialize graphene arguments

    get args from
        cls._args
        cls._required_args
    args is dict { arg_name: arg_type }
    or list [arg_name,] -> defaults to String
    """

    _args = {}
    _required_args = {}

    def __init_subclass__(cls, **kwargs):
        if isinstance(cls._args, dict):
            for key in cls._args:
                setattr(cls, key)
        elif isinstance(cls._args, list):
            for key in cls._args:
                cls._meta.arguments.update({key: graphene.String()})

        if isinstance(cls._required_args, dict):
            for key in cls._required_args:
                cls._meta.arguments.update(
                    {
                        key: graphene.Argument(
                            getattr(graphene, cls._required_args[key]), required=True
                        )
                    }
                )
        elif isinstance(cls._required_args, list):
            for key in cls._required_args:
                cls._meta.arguments.update({key: graphene.String(required=True)})
        return super().Field(*args, **kwargs)





import inspect
from functools import wraps
from strawberry_django_jwt.decorators import login_required as login_req
from strawberry_django_jwt.exceptions import PermissionDenied
from gqlauth.utils import g_user
from .constants import Messages
from .exceptions import WrongUsage, GraphQLAuthError




def login_required(fn):
    """
    If the user is registered
    """

    @wraps(fn)
    def wrapper(src, info, **kwargs):
        if g_user(info).is_authenticated:
            return fn(src, info, **kwargs)
        else:
            return src.output(success=False, errors=Messages.UNAUTHENTICATED)

    return wrapper


def verification_required(fn):
    """
    if the user was approved
    """

    @wraps(fn)
    @login_required
    def wrapper(src, info, **kwargs):
        if not g_user(info).status.verified:
            return src.output(success=False, errors=Messages.NOT_VERIFIED)
        return fn(src, info, **kwargs)

    return wrapper


def secondary_email_required(fn):
    @wraps(fn)
    @verification_required
    def wrapper(src, info, **kwargs):
        if not g_user(info).status.secondary_email:
            return src.output(success=False, errors=Messages.SECONDARY_EMAIL_REQUIRED)
        return fn(src, info, **kwargs)

    return wrapper


def password_confirmation_required(fn):
    @wraps(fn)
    def wrapper(src, info, **kwargs):
        try:
            password_arg = next(
                i for i in kwargs.keys() if i in ["password", "oldPassword"]
            )
            password = kwargs[password_arg]
        except Exception:
            raise WrongUsage(
                """
                @password_confirmation is supposed to be used on
                user with 'password' or 'old_password' field required.
                """
            )
        user = g_user(info)
        if user.check_password(password):
            return fn(src, info, **kwargs)
        errors = {password_arg: Messages.INVALID_PASSWORD}
        return src.output(success=False, errors=errors)

    return wrapper


def allowed_permissions(roles: list):
    """
    checks a list of roles if it applies to a user
    verification required by default.
    """

    def decorator(fn):
        @wraps(fn)
        @verification_required
        def wrapper(src, info, **kwargs):
            user = g_user(info)

        return wrapper

    return decorator



from django.contrib.auth import get_user_model
from uuid import UUID



# strawberry
import strawberry
from strawberry.django import auth, auto


# project
from gqlauth import models

USER_MODEL = get_user_model()

@strawberry.django.filters.filter(models.UserStatus)
class UserStatusFilter:
    verified: auto
    archived: auto
    secondary_email: auto


@strawberry.django.type(model=models.UserStatus, filters=UserStatusFilter)
class UserStatusType:
    verified: auto
    archived: auto
    secondary_email: auto


@strawberry.django.filters.filter(USER_MODEL)
class UserFilter:
    uuid: auto
    username: auto
    email: auto
    first_name: auto
    last_name: auto
    logentry_set: auto
    is_superuser: auto
    last_login: auto
    is_staff: auto
    is_active: auto
    date_joined: auto
    # connablealert_set: auto
    # groups: auto
    # user_permissions: auto
    status: UserStatusFilter


@strawberry.django.type(model=USER_MODEL, filters=UserFilter)
class UserType:
    uuid: auto
    username: auto
    email: auto
    first_name: auto
    last_name: auto
    logentry_set: auto
    is_superuser: auto
    last_login: auto
    is_staff: auto
    is_active: auto
    date_joined: auto
    # status: UserStatusType

    @strawberry.django.field
    def pk(self, info) -> int:
        return self.pk

    @strawberry.django.field
    def archived(self, info) -> bool:
        return self.status.archived

    @strawberry.django.field
    def verified(self, info) -> bool:
        return self.status.verified

    @strawberry.django.field
    def secondary_email(self, info) -> str:
        return self.status.secondary_email

    def get_queryset(self, queryset, info, **kwargs):
        return queryset.select_related("status")



from gqlauth.bases.mixins import (
    DynamicArgsMutationMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
)
from gqlauth.user.resolvers import (
    RegisterMixin,
    VerifyAccountMixin,
    ResendActivationEmailMixin,
    SendPasswordResetEmailMixin,
    PasswordSetMixin,
    PasswordResetMixin,
    ObtainJSONWebTokenMixin,
    ArchiveAccountMixin,
    DeleteAccountMixin,
    PasswordChangeMixin,
    UpdateAccountMixin,
    RefreshTokenMixin,
    VerifyTokenMixin,
    RevokeTokenMixin,
    SendSecondaryEmailActivationMixin,
    VerifySecondaryEmailMixin,
    SwapEmailsMixin,
    RemoveSecondaryEmailMixin,
)
class Register(
    RegisterMixin, DynamicArgsMixin, DynamicPayloadMixin, DynamicArgsMutationMixin
):
    __doc__ = RegisterMixin.__doc__

from strawberry_django_jwt.mutations import  Verify as VerifyParent
class VerifyAccount(
    VerifyAccountMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
    VerifyParent
):
    __doc__ = VerifyAccountMixin.__doc__



class ResendActivationEmail(
    ResendActivationEmailMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin
):
    __doc__ = ResendActivationEmailMixin.__doc__



class SendPasswordResetEmail(
    SendPasswordResetEmailMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin
):
    __doc__ = SendPasswordResetEmailMixin.__doc__



class SendSecondaryEmailActivation(
    SendSecondaryEmailActivationMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
):
    __doc__ = SendSecondaryEmailActivationMixin.__doc__



class VerifySecondaryEmail(
    VerifySecondaryEmailMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
):
    __doc__ = VerifySecondaryEmailMixin.__doc__



class SwapEmails(
    SwapEmailsMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
):
    __doc__ = SwapEmailsMixin.__doc__



class RemoveSecondaryEmail(
    RemoveSecondaryEmailMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
):
    __doc__ = RemoveSecondaryEmailMixin.__doc__



class PasswordSet(
    PasswordSetMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
):
    __doc__ = PasswordSetMixin.__doc__



class PasswordReset(
    PasswordResetMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
):
    __doc__ = PasswordResetMixin.__doc__

from strawberry_django_jwt.mutations import ObtainJSONWebToken as JwtObtainParent
class ObtainJSONWebToken(
    ObtainJSONWebTokenMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
    JwtObtainParent,
):
    __doc__ = ObtainJSONWebTokenMixin.__doc__




class ArchiveAccount(
    ArchiveAccountMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
):
    __doc__ = ArchiveAccountMixin.__doc__



class DeleteAccount(
    DeleteAccountMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
):
    __doc__ = DeleteAccountMixin.__doc__



class PasswordChange(
    PasswordChangeMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
    JwtObtainParent
):
    __doc__ = PasswordChangeMixin.__doc__



class UpdateAccount(
    UpdateAccountMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
):
    __doc__ = UpdateAccountMixin.__doc__


from strawberry_django_jwt.mutations import Verify as VerifyParent
class VerifyToken(
    VerifyTokenMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
    VerifyParent
):
    __doc__ = VerifyTokenMixin.__doc__



from strawberry_django_jwt.mutations import Refresh
class RefreshToken(
    RefreshTokenMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
    Refresh
):
    __doc__ = RefreshTokenMixin.__doc__


from strawberry_django_jwt.mutations import Revoke as RevokeParent

class RevokeToken(
    RevokeTokenMixin,
    DynamicArgsMixin,
    DynamicPayloadMixin,
    DynamicArgsMutationMixin,
    RevokeParent
):
    __doc__ = RevokeTokenMixin.__doc__




from django.utils.translation import gettext as _


class ErrorBase(Exception):
    default_message = None

    def __init__(self, message=None):
        if message is None:
            message = self.default_message

        super().__init__(message)


class WrongUsage(ErrorBase):
    """
    internal exception
    """
    default_message = _("Wrong usage, check your code!.")

class WrongInput(ErrorBase):
    default_message = _("Wrong Input received")




"""
Django settings for project project.

Generated by 'django-admin startproject' using Django 3.0.2.

For more information on this file, see
https://docs.djangoproject.com/en/3.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.0/ref/settings/
"""

import os

BASE_DIR = os.path.dirname(__file__)

SECRET_KEY = "2qxe8hr#pz3t#0gd$0)rh)3fxr3+j3(k3a!ao)6d&y71mmdxk2"

DEBUG = True

ALLOWED_HOSTS = []

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    'strawberry_django',
    'strawberry_django_jwt.refresh_token',
    'users',
    'gqlauth',  # NOTE authantication with jwt
]

AUTH_USER_MODEL = 'users.CustomUser'

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
]

ROOT_URLCONF = "urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]


DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

STATIC_URL = "/static/"


# custom settings start here


AUTHENTICATION_BACKENDS = [
    "gqlauth.backends.GraphQLAuthBackend",
    "django.contrib.auth.backends.ModelBackend",
]

GRAPHQL_JWT = {
    "JWT_VERIFY_EXPIRATION": True,
    "JWT_LONG_RUNNING_REFRESH_TOKEN": True,
    "JWT_ALLOW_ANY_CLASSES": [
        "gqlauth.mutations.Register",
        "gqlauth.mutations.VerifyAccount",
        "gqlauth.mutations.ResendActivationEmail",
        "gqlauth.mutations.SendPasswordResetEmail",
        "gqlauth.mutations.PasswordReset",
        "gqlauth.mutations.ObtainJSONWebToken",
        "gqlauth.mutations.VerifyToken",
        "gqlauth.mutations.RefreshToken",
        "gqlauth.mutations.RevokeToken",
        "gqlauth.mutations.VerifySecondaryEmail",
    ],
}
import gqlauth

EMAIL_BACKEND = "django.core.mail.backends.console.EmailBackend"

GQL_AUTH = {
    "EMAIL_ASYNC_TASK": "pseudo_async_email_support.pseudo_async_email_support"
}



from django.utils.translation import gettext as _


class Messages:
    INVALID_PASSWORD = [{"message": _("Invalid password."), "code": "invalid_password"}]
    UNAUTHENTICATED = [{"message": _("Unauthenticated."), "code": "unauthenticated"}]
    INVALID_TOKEN = [{"message": _("Invalid token."), "code": "invalid_token"}]
    EXPIRED_TOKEN = [{"message": _("Expired token."), "code": "expired_token"}]
    NO_SUFFICIENT_PERMISSIONS = [{"message": _("Expired token."), "code": "expired_token"}]
    
    # captcha messages
    CAPTCHA_VALID = [{"message": _("Captcha successfully validated!"), "code": "valid_captcha"}]
    CAPTCHA_INVALID = [{"message": _("Captcha wrong, try again."), "code": "invalid_captcha"}]
    CAPTCHA_MAX_RETRIES = [{"message": _("Maximum tries exceeded, please refresh the captcha."), "code": "max_retries_exceeded"}]
    CAPTCHA_EXPIRED = [{"message": _("Expired or not Existed captcha please refresh."), "code": "expired_captcha"}]

    ALREADY_VERIFIED = [
        {"message": _("Account already verified."), "code": "already_verified"}
    ]
    EMAIL_FAIL = [{"message": _("Failed to send email."), "code": "email_fail"}]
    INVALID_CREDENTIALS = [
        {
            "message": _("Please, enter valid credentials."),
            "code": "invalid_credentials",
        }
    ]
    NOT_VERIFIED = [
        {"message": _("Please verify your account."), "code": "not_verified"}
    ]
    NOT_VERIFIED_PASSWORD_RESET = [
        {
            "message": _("Verify your account. A new verification email was sent."),
            "code": "not_verified",
        }
    ]
    EMAIL_IN_USE = [
        {"message": _("A user with that email already exists."), "code": "unique"}
    ]
    SECONDARY_EMAIL_REQUIRED = [
        {
            "message": _("You need to setup a secondary email to proceed."),
            "code": "secondary_email_required",
        }
    ]
    PASSWORD_ALREADY_SET = [
        {
            "message": _("Password already set for account."),
            "code": "password_already_set",
        }
    ]


class TokenAction(object):
    ACTIVATION = "activation"
    PASSWORD_RESET = "password_reset"
    ACTIVATION_SECONDARY_EMAIL = "activation_secondary_email"
    PASSWORD_SET = "password_set"



from gqlauth.utils import camelize
import strawberry
from gqlauth.bases.exceptions import WrongUsage




def serialize_excpected_error(errors):
    if isinstance(errors, dict):
        if errors.get("__all__", False):
            errors["non_field_errors"] = errors.pop("__all__")
        return camelize(errors)
    elif isinstance(errors, list):
        return {"nonFieldErrors": errors}
    raise WrongUsage("`errors` must be list or dict!")


@strawberry.scalar(name='ExpectedError',
                   serialize=lambda value: serialize_excpected_error(value),
                   parse_value=lambda value: ExpectedErrorType(value)
                   )
class ExpectedErrorType:
    """
    Errors messages and codes mapped to
    fields or non fields errors.
    Example:
    {
        field_name: [
            {
                "message": "error message",
                "code": "error_code"
            }
        ],
        other_field: [
            {
                "message": "error message",
                "code": "error_code"
            }
        ],
        nonFieldErrors: [
            {
                "message": "error message",
                "code": "error_code"
            }
        ]
    }
    """
    pass




from strawberry_django_jwt.backends import JSONWebTokenBackend
from strawberry_django_jwt.shortcuts import get_user_by_token
from strawberry_django_jwt.utils import get_credentials
from strawberry_django_jwt.exceptions import JSONWebTokenError, JSONWebTokenExpired


class GraphQLAuthBackend(JSONWebTokenBackend):
    """
    Only difference from the original backend
    is it does not raise when fail on get_user_by_token
    preventing of raise when client send token to a
    mutation that does not requery login but is not on
    allow any settings.

    Main advantage is to let the mutation handle the
    unauthentication error. Intead of an actual error,
    we can return e.g. success=False errors=Unauthenticated
    """

    def authenticate(self, request=None, **kwargs):
        # if it isnt a request or settings dosnt require a _jwt_token
        if request is None or getattr(request, "_jwt_token_auth", False):
            return None



        try:  # +++
            token = get_credentials(request, **kwargs)
            if token is not None:
                return get_user_by_token(token, request)
        except JSONWebTokenError:  # +++
            pass  # +++

        except JSONWebTokenExpired:
            ...
        return None



import logging
from gqlauth.bases.mixins import DynamicRelayMutationMixin, DynamicInputMixin
from gqlauth.administrator.mixins import VerifyUserMixin

logging.getLogger(name=__name__)



class VerifyUser(DynamicRelayMutationMixin, DynamicInputMixin, VerifyUserMixin):
    ...



from django.db import models
from django.contrib.auth.models import AbstractUser
from django.utils.translation import gettext_lazy as _
import uuid


class CustomUser(AbstractUser):
    uuid = models.UUIDField(default=uuid.uuid4, editable=False)
    email = models.EmailField(blank=False, unique=True, max_length=254, verbose_name=_("email address"))

    USERNAME_FIELD = "username"  # e.g: "username", "email"
    EMAIL_FIELD = "email"  # e.g: "email", "primary_email"




from django.contrib import admin

from django.apps import apps

# Register your models here.
app = apps.get_app_config('gqlauth')

for model_name, model in app.models.items():
    admin.site.register(model)



from base64 import b64encode, b64decode
import typing
import strawberry
from PIL.Image import Image

image = strawberry.scalar(
    typing.NewType('image', Image),
    serialize=lambda v: b64encode(v).decode('ascii'),
    parse_value=lambda v: b64decode(v),
)
        





"""
Settings for gqlauth are all namespaced in the GQL_AUTH setting.
For example your project's `settings.py` file might look like this:
GQL_AUTH = {
    "LOGIN_OPTIONAL_FIELDS": ["email", "username"],
    "SEND_ACTIVATION_EMAIL": True,
}
This module provides the `gqlauth_settings` object, that is used to access
strawberry settings, checking for user settings first, then falling
back to the defaults.
"""

from django.conf import settings as django_settings
from django.test.signals import setting_changed

from datetime import timedelta

# Copied shamelessly from Graphene / Django REST Framework

DEFAULTS = {
    # if allow logging in without verification,
    # the register mutation will return a token
    "ALLOW_LOGIN_NOT_VERIFIED": False,
    # mutation fields options
    "LOGIN_OPTIONAL_FIELDS": [],
    "LOGIN_REQUIRE_CAPTCHA": True,
    "LOGIN_REQUIRED_FIELDS": ['username', 'password'],
    # required fields on register, plus password1 and password2,
    # can be a dict like UPDATE_MUTATION_FIELDS setting
    "REGISTER_MUTATION_FIELDS": ["email", "username"],
    "REGISTER_MUTATION_FIELDS_OPTIONAL": [],
    "REGISTER_REQUIRE_CAPTCHA": True,
    "CAPTCHA_TEXT_FACTORY": None,
    # captcha stuff
    "CAPTCHA_EXPIRATION_DELTA": timedelta(seconds=120),
    "CAPTCHA_MAX_RETRIES": 5,
    # optional fields on update account, can be list of fields
    "UPDATE_MUTATION_FIELDS": {"first_name": str, "last_name": str},
    # tokens
    "EXPIRATION_ACTIVATION_TOKEN": timedelta(days=7),
    "EXPIRATION_PASSWORD_RESET_TOKEN": timedelta(hours=1),
    "EXPIRATION_SECONDARY_EMAIL_ACTIVATION_TOKEN": timedelta(hours=1),
    "EXPIRATION_PASSWORD_SET_TOKEN": timedelta(hours=1),
    # email stuff
    "EMAIL_FROM": getattr(django_settings, "DEFAULT_FROM_EMAIL", "test@email.com"),
    "SEND_ACTIVATION_EMAIL": False,
    # client: example.com/activate/token
    "ACTIVATION_PATH_ON_EMAIL": "activate",
    "ACTIVATION_SECONDARY_EMAIL_PATH_ON_EMAIL": "activate",
    # client: example.com/password-set/token
    "PASSWORD_SET_PATH_ON_EMAIL": "password-set",
    # client: example.com/password-reset/token
    "PASSWORD_RESET_PATH_ON_EMAIL": "password-reset",
    # email subjects templates
    "EMAIL_SUBJECT_ACTIVATION": "email/activation_subject.txt",
    "EMAIL_SUBJECT_ACTIVATION_RESEND": "email/activation_subject.txt",
    "EMAIL_SUBJECT_SECONDARY_EMAIL_ACTIVATION": "email/activation_subject.txt",
    "EMAIL_SUBJECT_PASSWORD_SET": "email/password_set_subject.txt",
    "EMAIL_SUBJECT_PASSWORD_RESET": "email/password_reset_subject.txt",
    # email templates
    "EMAIL_TEMPLATE_ACTIVATION": "email/activation_email.html",
    "EMAIL_TEMPLATE_ACTIVATION_RESEND": "email/activation_email.html",
    "EMAIL_TEMPLATE_SECONDARY_EMAIL_ACTIVATION": "email/activation_email.html",
    "EMAIL_TEMPLATE_PASSWORD_SET": "email/password_set_email.html",
    "EMAIL_TEMPLATE_PASSWORD_RESET": "email/password_reset_email.html",
    "EMAIL_TEMPLATE_VARIABLES": {},
    # query stuff
    "USER_NODE_EXCLUDE_FIELDS": ["password", "is_superuser"],
    "USER_NODE_FILTER_FIELDS": {
        "email": ["exact"],
        "username": ["exact", "icontains", "istartswith"],
        "is_active": ["exact"],
        "status__archived": ["exact"],
        "status__verified": ["exact"],
        "status__secondary_email": ["exact"],
    },
    # turn is_active to False instead
    "ALLOW_DELETE_ACCOUNT": False,
    # string path for email function wrapper, see the testproject example
    "EMAIL_ASYNC_TASK": False,
    # mutation error type
    "CUSTOM_ERROR_TYPE": None,
    # registration with no password
    "ALLOW_PASSWORDLESS_REGISTRATION": False,
    "SEND_PASSWORD_SET_EMAIL": False,
}


class GraphQLAuthSettings(object):
    """
    A settings object, that allows API settings to be accessed as properties.
    For example:
        from gqlauth.settings import settings
        print(settings)
    """

    def __init__(self, user_settings=None, defaults=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(django_settings, "GQL_AUTH", {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid gqlauth setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Cache the result
        setattr(self, attr, val)
        return val


gqlauth_settings = GraphQLAuthSettings(None, DEFAULTS)


def reload_gqlauth_settings(*args, **kwargs):
    global gqlauth_settings
    setting, value = kwargs["setting"], kwargs["value"]
    if setting == "GQL_AUTH":
        gqlauth_settings = GraphQLAuthSettings(value, DEFAULTS)


setting_changed.connect(reload_gqlauth_settings)



import strawberry
from gqlauth.models import Captcha
from gqlauth.types import CaptchaType
from gqlauth.user.mutations import UserMutations
from gqlauth.user.queries import UserQueries
from strawberry.tools import merge_types

@strawberry.type
class CaptchaMutation:
    @strawberry.mutation
    def captcha(self, info) -> CaptchaType:
        return Captcha.create_captcha()


AuthQueries = merge_types('Auth', (UserQueries,))
AuthMutations = merge_types('AuthMutations', (CaptchaMutation, UserMutations))




# Generated by Django 4.0.3 on 2022-04-04 07:15

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Captcha',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False)),
                ('text', models.CharField(editable=False, max_length=50)),
                ('insert_time', models.DateTimeField(auto_now_add=True)),
                ('tries', models.IntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name='UserStatus',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('verified', models.BooleanField(default=False)),
                ('archived', models.BooleanField(default=False)),
                ('secondary_email', models.EmailField(blank=True, max_length=254, null=True)),
            ],
        ),
    ]



from django.apps import AppConfig


class GqlAuthConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = "gqlauth"
    verbose_name = "GraphQL Auth"

    def ready(self):
        from . import signals



from uuid import UUID
import strawberry_django
import strawberry
from strawberry_django import auto

from gqlauth import models
from gqlauth.scalars import image


@strawberry_django.type(model=models.Captcha)
class CaptchaType:
    uuid: UUID
    @strawberry.field
    def image(self) -> image:
        return self.as_bytes()



from django.utils.translation import gettext as _
from gqlauth.bases.exceptions import ErrorBase


class GraphQLAuthError(ErrorBase):
    def __init__(self, message=None):
        if message is None:
            message = " ".join([_("Authorization error:"), self.default_message])

        super().__init__(message)


class UserAlreadyVerified(GraphQLAuthError):
    default_message = _("User already verified.")


class InvalidCredentials(GraphQLAuthError):
    default_message = _("Invalid credentials.")


class UserNotVerified(GraphQLAuthError):
    default_message = _("User is not verified.")


class EmailAlreadyInUse(GraphQLAuthError):
    default_message = _("This email is already in use.")


class TokenScopeError(GraphQLAuthError):
    default_message = _("This token if for something else.")


class PasswordAlreadySetError(GraphQLAuthError):
    default_message = _("Password already set for account.")


class WrongUsage(GraphQLAuthError):
    """
    internal exception
    """
    default_message = _("Wrong usage, check your code!.")


class NoSufficiantPermissions(GraphQLAuthError):
    default_message = _("User is not allowed for this content")



import typing
from typing import Union
import inspect
import dataclasses
from django.contrib.auth import get_user_model
from django.utils.module_loading import import_string

import strawberry
from strawberry.utils.str_converters import to_camel_case
from gqlauth.settings import gqlauth_settings as app_settings
from gqlauth.utils import create_strawberry_argument
from gqlauth.utils import list_to_dict, hide_args_kwargs
from gqlauth.bases.exceptions import (
    WrongUsage
)
from gqlauth.bases.interfaces import OutputInterface
from gqlauth.bases.scalars import ExpectedErrorType

UserModel = get_user_model()
if app_settings.EMAIL_ASYNC_TASK and isinstance(app_settings.EMAIL_ASYNC_TASK, str):
    async_email_func = import_string(app_settings.EMAIL_ASYNC_TASK)
else:
    async_email_func = None


def make_dataclass_helper(required: dict | list, non_required: dict | list, camelize=True):
    res_req = []
    res_non_req = []

    if isinstance(required, dict):
        if camelize:
            for key in required:
                res_req.append((to_camel_case(key), required[key]))
        else:
            for key in required:
                res_req.append((key, required[key]))

    elif isinstance(required, list):
        if camelize:
            for key in required:
                res_req.append((to_camel_case(key), str))

        else:
            for key in required:
                res_req.append((key, str))

    if isinstance(non_required, dict):
        if camelize:
            for key in non_required:
                res_non_req.append(
                    (to_camel_case(key), typing.Optional[non_required[key]], dataclasses.field(default=None)))
        else:
            for key in non_required:
                res_non_req.append((key, typing.Optional[non_required[key]], dataclasses.field(default=None)))

    elif isinstance(non_required, list):
        if camelize:
            for key in non_required:
                res_non_req.append((to_camel_case(key), typing.Optional[str], dataclasses.field(default=None)))
        else:
            for key in non_required:
                res_non_req.append((key, typing.Optional[str], dataclasses.field(default=None)))

    return res_req + res_non_req


def is_optional(field):
    return typing.get_origin(field) is Union and \
           type(None) in typing.get_args(field)


class DynamicInputMixin:
    """
    A class that knows how to initialize strawberry input

    as dict { arg name: arg type }
    as list [arg_name,] -> defaults to String

    ### usage:
        class SomeMutation(DynamicInputMixin, MutationMixin):
                class _meta:
                    _inputs = [some, non, required, fields]
                    _required_inputs = {some:str, required:int, fields:list}

                def resolve_mutation(self, input_):
                    logic...

    """

    def __init_subclass__(cls, **kwargs):
        _inputs = getattr(cls._meta, '_inputs', [])
        _required_inputs = getattr(cls._meta, '_required_inputs', [])

        if _inputs or _required_inputs:
            if not isinstance(_inputs, dict | list) and _inputs:
                raise WrongUsage(f'dynamic inputs can be list or dict not{type(_inputs)}')
            if not isinstance(_required_inputs, dict | list) and _required_inputs:
                raise WrongUsage(f'dynamic required inputs can be list or dict not{type(_required_inputs)}')

        parent_resolver_name = getattr(cls._meta, '_parent_resolver_name', None)
        if parent_resolver_name:
            _inputs = list_to_dict(_inputs)
            _required_inputs = list_to_dict(_required_inputs)

            parent_resolver = getattr(cls, parent_resolver_name, None)
            for name, ret_type in parent_resolver.base_resolver.annotations.items():
                if name != 'return':
                    if is_optional(ret_type):
                        _inputs[name] = typing.get_args(ret_type)[0]
                    else:
                        _required_inputs[name] = ret_type

        if _inputs or _required_inputs:
            dc = dataclasses.make_dataclass(f'{cls.__name__}Input',
                                            [
                                                *make_dataclass_helper(
                                                    _required_inputs, _inputs),
                                            ]
                                            )

            inputs = strawberry.input(dc)
            setattr(cls._meta, 'inputs', inputs)
        else:
            setattr(cls._meta, 'inputs', None)
        super().__init_subclass__()

    def Field(self):
        raise NotImplemented


class DynamicArgsMixin:
    """
    A class that knows how to initialize strawberry arguments

    as dict { arg name: arg type }
    as list [arg_name,]  defaults to String

    ### usage:
        class SomeMutation(DynamicInputMixin, MutationMixin):
                class _meta:
                    _inputs = [some, non, required, fields]
                    _required_inputs = {some:str, required:int, fields:list}

                def resolve_mutation(self, input_):
                    logic...

    """

    def __init_subclass__(cls, **kwargs):
        _inputs = getattr(cls._meta, '_inputs', [])
        _required_inputs = getattr(cls._meta, '_required_inputs', [])

        if _inputs or _required_inputs:
            if not isinstance(_inputs, dict | list) and _inputs:
                raise WrongUsage(f'dynamic inputs can be list or dict not{type(_inputs)}')
            if not isinstance(_required_inputs, dict | list) and _required_inputs:
                raise WrongUsage(f'dynamic required inputs can be list or dict not{type(_required_inputs)}')

        parent_resolver_name = getattr(cls._meta, '_parent_resolver_name', None)
        if parent_resolver_name:
            _inputs = list_to_dict(_inputs)
            _required_inputs = list_to_dict(_required_inputs)

            parent_resolver = getattr(cls, parent_resolver_name, None)
            for name, ret_type in parent_resolver.base_resolver.annotations.items():
                if name != 'return':
                    if is_optional(ret_type):
                        _inputs[name] = typing.get_args(ret_type)[0]
                    else:
                        _required_inputs[name] = ret_type

        if _inputs or _required_inputs:

            args = make_dataclass_helper(
                _required_inputs, _inputs)

            setattr(cls._meta, 'args', args)
        else:
            setattr(cls._meta, 'args', [])
        super().__init_subclass__()

    @property
    def Field(self):
        raise NotImplemented(
            'This mimxin has to be mixed with either `DynamicRelayMutationMixin`,'
            ' or `DynamicDefaultMutationMixin`')


class DynamicPayloadMixin:
    """
        A class that knows how to initialize strawberry
        payload from a list or dict
        and to merge parent payload to the given payload

        as dict { payload_name: payload_type }
        as list [input_name,] -> defaults to String

        #usage:
            class SomeMutation(DynamicInputMixin, OutputMixin, MutationMixin):
                    _outputs = [some, non, required, fields]
                    _required_outputs = {some:str, required:int, fields:list}

                    def resolve_mutation(self, input_):
                        logic...
        """

    def __init_subclass__(cls, **kwargs):
        _outputs = getattr(cls._meta, '_outputs', [])
        _required_outputs = getattr(cls._meta, '_required_outputs', [])
        if _outputs or _required_outputs:
            if not isinstance(_outputs, dict | list | None) and _outputs:
                raise WrongUsage(f'dynamic outputs can be list or dict not{type(_outputs)}')
            if not isinstance(_required_outputs, dict | list | None) and _required_outputs:
                raise WrongUsage(f'dynamic required outputs can be list or dict not{type(_required_outputs)}')

        parent_resolver_name = getattr(cls._meta, '_parent_resolver_name', None)
        if parent_resolver_name:
            parent_resolver = getattr(cls, parent_resolver_name, None)
            new_dict = {}
            if isinstance(_outputs, list):
                # must create a dict for non str types
                for output in _outputs:
                    new_dict[output] = str
                _outputs = new_dict

            _outputs[parent_resolver_name + 'Payload'] = parent_resolver.type

        if _outputs or _required_outputs:
            dc = dataclasses.make_dataclass(f'{cls.__name__}Outputs',
                                            [
                                                ('success', bool),
                                                *make_dataclass_helper(
                                                    _required_outputs, _outputs),
                                                ('errors', typing.Optional[ExpectedErrorType],
                                                 dataclasses.field(default=None)),
                                            ],
                                            namespace={
                                                'get': lambda self, attr: self.__getattribute__(attr)
                                            }
                                            )
            outputs = strawberry.type(dc)
        else:
            outputs = strawberry.type(OutputInterface)

        setattr(cls, 'output', outputs)
        super().__init_subclass__()

    @property
    def Field(self):
        raise NotImplemented('This mimxin has to be mixed with either `DynamicRelayMutationMixin`,'
                             ' or `DynamicDefaultMutationMixin`')


class DynamicRelayMutationMixin:
    def __init_subclass__(cls, **kwargs):
        @strawberry.mutation(description=cls.__doc__)
        def Field(info, input_: cls._meta.inputs) -> cls.output:
            arguments = {f.name: getattr(input_, f.name) for f in dataclasses.fields(input_)}
            return cls.resolve_mutation(info, **arguments)

        if Field.arguments[0].type is None: Field.arguments.pop(0)  # if no input
        setattr(cls, 'Field', Field)

    def resolve_mutation(self, info, **kwargs):
        raise NotImplemented('This has method has to be implemented by the logic delegated mixin')


class DynamicArgsMutationMixin:
    def __init_subclass__(cls, **kwargs):
        def Field(info, **kwargs) -> cls.output:
            return cls.resolve_mutation(info, **kwargs)

        Field = hide_args_kwargs(Field)
        Field = strawberry.mutation(Field, description=cls.__doc__)

        for arg_tuple in cls._meta.args:
            arg = create_strawberry_argument(arg_tuple[0], arg_tuple[0], arg_tuple[1])
            Field.arguments.append(arg)
        setattr(cls, 'Field', Field)

    def resolve_mutation(self, info, *args, **kwargs):
        raise NotImplemented('This has method has to be implemented by the logic delegated mixin')






from django.contrib.auth.forms import UserCreationForm, UserChangeForm, UsernameField, PasswordChangeForm
from django.contrib.auth import get_user_model
from django import forms

from .utils import flat_dict
from gqlauth.settings import gqlauth_settings as app_settings


class RegisterForm(UserCreationForm):
    email = forms.EmailField(required=True if 'email' in app_settings.REGISTER_MUTATION_FIELDS else False)

    class Meta:
        model = get_user_model()
        fields = flat_dict(app_settings.REGISTER_MUTATION_FIELDS) + flat_dict(
            app_settings.REGISTER_MUTATION_FIELDS_OPTIONAL
        )


class PasswordChangeFormGql(PasswordChangeForm):
    field_order = ['oldPassword', 'newPassword1', 'newPassword2']

class EmailForm(forms.Form):
    email = forms.EmailField(max_length=254)


class CustomUsernameField(UsernameField):
    required = False


class UpdateAccountForm(UserChangeForm):
    class Meta:
        model = get_user_model()
        fields = flat_dict(app_settings.UPDATE_MUTATION_FIELDS)
        field_classes = {"username": CustomUsernameField}


class PasswordLessRegisterForm(UserCreationForm):
    """
    A RegisterForm with optional password inputs.
    """

    class Meta:
        model = get_user_model()
        fields = flat_dict(app_settings.REGISTER_MUTATION_FIELDS) + flat_dict(
            app_settings.REGISTER_MUTATION_FIELDS_OPTIONAL
        )

    def __init__(self, *args, **kwargs):
        super(PasswordLessRegisterForm, self).__init__(*args, **kwargs)
        self.fields["password1"].required = False
        self.fields["password2"].required = False

    def save(self, commit=True):
        user = super().save(commit=False)
        user.set_unusable_password()
        if commit:
            user.save()
        return user



import strawberry
from strawberry.tools import merge_types

from typing import List

from gqlauth.user import arg_mutations
from gqlauth.user.types import UserType, UserFilter



@strawberry.type
class UserQueries:
    user: UserType = strawberry.django.field()
    users: List[UserType] = strawberry.django.field(filters=UserFilter)

@strawberry.type
class AuthMutation:
    token_auth = arg_mutations.ObtainJSONWebToken.Field
    verify_token = arg_mutations.VerifyToken.Field
    refresh_token = arg_mutations.RefreshToken.Field
    revoke_token = arg_mutations.RevokeToken.Field
    register = arg_mutations.Register.Field
    verify_account = arg_mutations.VerifyAccount.Field
    update_account = arg_mutations.UpdateAccount.Field
    resend_activation_email = arg_mutations.ResendActivationEmail.Field
    archive_account = arg_mutations.ArchiveAccount.Field
    delete_account = arg_mutations.DeleteAccount.Field
    password_change = arg_mutations.PasswordChange.Field
    send_password_reset_email = arg_mutations.SendPasswordResetEmail.Field
    password_reset = arg_mutations.PasswordReset.Field
    password_set = arg_mutations.PasswordSet.Field
    verify_secondary_email = arg_mutations.VerifySecondaryEmail.Field
    swap_emails = arg_mutations.SwapEmails.Field
    remove_secondary_email = arg_mutations.RemoveSecondaryEmail.Field
    send_secondary_email_activation = arg_mutations.SendSecondaryEmailActivation.Field


Query = merge_types('RootQuery', (UserQueries,))

Mutation = merge_types('RootMutation', (AuthMutation,))

schema = strawberry.Schema(query=Query, mutation=Mutation)



from . import relay
from . import arg_mutations
import strawberry


@strawberry.type
class UserMutations:
    register = relay.Register.Field
    login = relay.ObtainJSONWebToken.Field
    update_password = relay.PasswordChange.Field
    update_account_relay = relay.UpdateAccount.Field
    verify_token = relay.VerifyToken.Field
    refresh_token = relay.RefreshToken.Field
    revoke_token = relay.RevokeToken.Field
    update_account = arg_mutations.UpdateAccount.Field




#!/usr/bin/env python

import io
import os
import re
from collections import OrderedDict

from setuptools import find_packages, setup


def get_version(package):
    with io.open(os.path.join(package, "__init__.py")) as f:
        pattern = r'^__version__ = [\'"]([^\'"]*)[\'"]'
        return re.search(pattern, f.read(), re.MULTILINE).group(1)


tests_require = [
    "pytest>=3.6.3",
    "pytest-cov>=2.4.0",
    "pytest-django>=3.1.2",
    "coveralls",
]

dev_requires = ["black==19.3b0", "flake8==3.7.7"] + tests_require

setup(
    name="strawberry-django-auth",
    version=get_version("gqlauth"),
    license="MIT",
    description="Graphql and relay authentication with Graphene for Django.",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    author="nir-benlulu",
    author_email="nrbnlulu@gmail.com",
    maintainer="nir benlulu",
    url="https://github.com/nrbnlulu/strawberry-django-auth",
    project_urls=OrderedDict(
        (
            ("Documentation", "https://strawberry-django-auth.readthedocs.io/en/latest/"),
            ("Issues", "https://github.com/nrbnlulu/strawberry-django-auth/issues"),
        )
    ),
    packages=find_packages(exclude=["tests*"]),
    install_requires=[
        "Django>=2.2.0",
        "strawberry-django-jwt==0.2.0",
        "strawberry-graphql-django==0.2.5",
        "strawberry-graphql==0.104.3",
        "PyJWT==2.3.0",
        "Faker==13.3.4",
        "Pillow==9.1.0",
    ],

    tests_require=tests_require,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
    ],
    keywords="api graphql rest relay strawberry auth",
    zip_safe=False,
    include_package_data=True,
    extras_require={"test": tests_require, "dev": dev_requires},
)



from django.db.models.signals import post_save
from django.conf import settings as django_settings

from django.dispatch import Signal, receiver


@receiver(post_save, sender=django_settings.AUTH_USER_MODEL)
def create_user_status(sender, instance, created, **kwargs):
    if created:
        from .models import UserStatus

        UserStatus._default_manager.get_or_create(user=instance)
        

user_registered = Signal()
user_verified = Signal()
